/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Messages queues module.
 * 
 * This module provides the support functions for queueing and picking messages.
 * See queues.h for more information.
 *
 * The threads that call these functions must be in the cancellation state PTHREAD_CANCEL_ENABLE and type PTHREAD_CANCEL_DEFERRED.
 * This is the default state and type on thread creation.
 *
 * In order to destroy properly a queue, the application must:
 *  -> shutdown any process that can add into the queue first.
 *  -> pthread_cancel any thread that could be waiting on the queue.
 *  -> consume any message that is in the queue, using meq_tryget.
 *  -> then destroy the queue using meq_del.
 */

#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "waaad-internal.h"

/********************************************************************************************************/

/* A list item */
typedef struct _li_t_ {
	struct _li_t_ 	* next;
	struct _li_t_ 	* prev;
	msg_t		* msg;
} _li_t;

/* Initialize a list element */
#define LI_init( _li, _msg ) {	\
	(_li)->next = (_li);	\
	(_li)->prev = (_li);	\
	(_li)->msg  = (_msg);	\
}

/* Test if a list is empty */
#define LI_isempty( _li ) ((_li)->next == (_li))

/* Unlink an element */
#define LI_unlink( _li_ ) {			\
	(_li_)->prev->next = (_li_)->next;	\
	(_li_)->next->prev = (_li_)->prev;	\
	(_li_)->next = (_li_);			\
	(_li_)->prev = (_li_);			\
}

/* Add an element in a list at the end */
#define LI_push( _sentinel, _item) {		\
	(_item)->prev = (_sentinel)->prev;	\
	(_item)->next = (_sentinel);		\
	(_sentinel)->prev->next = (_item);	\
	(_sentinel)->prev = (_item);		\
}

/* Remove an element from the head of a list */
#define LI_pop( _sentinel, _item) {		\
	(_item) = (_sentinel)->next;		\
	assert((_item) != (_sentinel));		\
	LI_unlink(_item);			\
}
	

/* Internal representation of a queue */
typedef struct {
	int		eyec;	/* An eye catcher, also used to check a queue is valid. _MEQ_EYEC */
	int		count;	/* number of objects in the list */
	int		thrs;	/* number of threads waiting in the list */
	_li_t		list;	/* sentinel for the list */
	pthread_mutex_t	mtx;	/* Mutex protecting this queue */
	pthread_cond_t	cond;	/* condition variable of the list */
} _meq_t;

/* The eye catcher value */
#define _MEQ_EYEC	0xe7ec1130

/* Macro to cast */
#define _Q( _ptr ) ((_meq_t *)(_ptr))

/* Macro to check a pointer */
#define CHECK_QUEUE( _queue ) (( (_queue) != NULL) && ( _Q(_queue)->eyec == _MEQ_EYEC) )


/* The global queues pointers are actually declared here */
meq_t * g_meq_incoming = NULL;	/* Received messages */
meq_t * g_meq_outgoing = NULL;	/* Messages to send */
meq_t * g_meq_local = NULL;	/* Messages for local handling */

/********************************************************************************************************/

/* Initialize the module */
int meq_init ( void )
{
	int ret = 0;
	
	TRACE_ENTRY( "" );
	
	/* Initialize the global queues */
	ret = meq_new(&g_meq_incoming);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to initialize a queue: %s", strerror(ret));
		return ret;
	}
	ret = meq_new(&g_meq_outgoing);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to initialize a queue: %s", strerror(ret));
		return ret;
	}
	ret = meq_new(&g_meq_local);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to initialize a queue: %s", strerror(ret));
		return ret;
	}
	
	return 0;
}

/* End of the module */
int meq_fini ( void )
{
	int ret = 0;
	
	TRACE_ENTRY( "" );
	
	ret = meq_del(&g_meq_incoming);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to delete a queue: %s", strerror(ret));
		return ret;
	}
	ret = meq_del(&g_meq_outgoing);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to delete a queue: %s", strerror(ret));
		return ret;
	}
	ret = meq_del(&g_meq_local);
	if (ret != 0) {
		TRACE_DEBUG(INFO, "Failed to delete a queue: %s", strerror(ret));
		return ret;
	}
	return 0;
}

/* Create a new queue */
int meq_new ( meq_t ** queue )
{
	int ret = 0;
	_meq_t * new;
	
	TRACE_ENTRY( "%p", queue );
	if (!queue) {
		TRACE_DEBUG(INFO, "Received NULL. EINVAL");
		return EINVAL;
	}
	
	/* Create a new object */
	new = (_meq_t *) malloc (sizeof (_meq_t) );
	if (new == NULL) {
		log_error("Memory allocation failed: %s\n", strerror(errno));
		TRACE_DEBUG(INFO, "Malloc failed");
		return ENOMEM;
	}
	
	/* Initialize the content */
	memset(new, 0, sizeof(_meq_t));
	
	new->eyec = _MEQ_EYEC;
	LI_init(&new->list, NULL);
	
	ret = pthread_mutex_init(&new->mtx, NULL);
	if (ret != 0) {
		log_error("Mutex initialization failed: %s\n", strerror(ret));
		TRACE_DEBUG(INFO, "pthread_mutex_init failed");
		free(new);
		return ret;
	}
		
	ret = pthread_cond_init(&new->cond, NULL);
	if (ret != 0) {
		log_error("Condition variable initialization failed: %s\n", strerror(ret));
		TRACE_DEBUG(INFO, "pthread_cond_init failed");
		free(new);
		return ret;
	}
	
	/* We're done */
	*queue = (meq_t *) new;
	return 0;
}

/* Delete a queue */
int meq_del ( meq_t ** queue )
{
	int ret = 0;
	
	TRACE_ENTRY( "%p", queue );

	if ( ! CHECK_QUEUE( *queue ) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	ret = pthread_mutex_lock( &_Q(*queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	if ((_Q(*queue)->count != 0) || (_Q(*queue)->thrs != 0)) {
		TRACE_DEBUG(INFO, "The queue count is %d, thread count %d, cannot destroy", _Q(*queue)->count, _Q(*queue)->thrs);
		ret = pthread_mutex_unlock( &_Q(*queue)->mtx );
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_mutex_unlock failed (ignored): %s", strerror(ret));
		}
		return EINVAL;
	}
	
	/* sanity check */
	assert(LI_isempty(&_Q(*queue)->list));
	
	/* Ok, now invalidate the queue */
	_Q(*queue)->eyec = 0xdead;
	
	/* And destroy it */
	ret = pthread_mutex_unlock( &_Q(*queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	ret = pthread_cond_destroy( &_Q(*queue)->cond );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_cond_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	ret = pthread_mutex_destroy( &_Q(*queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_destroy failed: %s", strerror(ret));
		return ret;
	}
	
	free(*queue);
	*queue = NULL;
	return 0;
}

/* Retrieve the number of messages of a queue */
int meq_length ( meq_t * queue, int * length )
{
	int ret = 0;
	
	TRACE_ENTRY( "%p %p", queue, length );
	
	/* Check the parameters */
	if ( (! CHECK_QUEUE( queue )) || (length == NULL) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* lock the queue */
	ret = pthread_mutex_lock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Retrieve the count */
	*length = _Q(queue)->count;
	
	/* Unlock */
	ret = pthread_mutex_unlock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Done */
	return 0;
}

/* Enqueue a message */
int meq_post ( meq_t * queue, msg_t * msg )
{
	_li_t * new;
	int ret = 0;
	
	TRACE_ENTRY( "%p %p", queue, msg );
	
	/* Check the parameters */
	if ( (! CHECK_QUEUE( queue )) || (msg == NULL) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Create a new list item */
	new = (_li_t *) malloc (sizeof (_li_t));
	if (new == NULL) {
		log_error("Memory allocation failed: %s", strerror(errno));
		TRACE_DEBUG(INFO, "malloc failed");
		return ENOMEM;
	}
	
	LI_init(new, msg);
	
	/* lock the queue */
	ret = pthread_mutex_lock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Add the new message */
	LI_push( &_Q(queue)->list, new);
	_Q(queue)->count++;
	
	/* Signal if threads are asleep */
	if (_Q(queue)->thrs > 0) {
		ret = pthread_cond_signal(&_Q(queue)->cond);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_cond_signal failed: %s", strerror(ret));
			return ret;
		}
	}
	
	/* Unlock */
	ret = pthread_mutex_unlock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Done */
	return 0;
}

/* Retrieve a message, if any */
int meq_tryget ( meq_t * queue, msg_t ** msg )
{
	_li_t * li;
	int ret = 0;
	int wouldblock = 0;
	
	TRACE_ENTRY( "%p %p", queue, msg );
	
	/* Check the parameters */
	if ( (! CHECK_QUEUE( queue )) || (msg == NULL) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* lock the queue */
	ret = pthread_mutex_lock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Check queue status */
	if (_Q(queue)->count > 0) {
		/* There are messages in the queue, so pick the first one */
		LI_pop( &_Q(queue)->list, li );
		_Q(queue)->count--;
		*msg = li->msg;
		free(li);
	} else {
		wouldblock = 1;
		*msg = NULL;
	}
		
	/* Unlock */
	ret = pthread_mutex_unlock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Done */
	return wouldblock ? EWOULDBLOCK : 0;
}

/* This handler is called when a thread is blocked on a queue, and cancelled */
static void _meq_cleanup(void * queue)
{
	int ret;
	
	TRACE_ENTRY( "%p", queue );
	
	/* Check the parameter */
	if ( ! CHECK_QUEUE( queue )) {
		TRACE_DEBUG(INFO, "Invalid queue, skipping handler");
		return;
	}
	
	/* The thread has been cancelled, therefore it does not wait on the queue anymore */
	_Q(queue)->thrs--;
	
	/* Now unlock the queue, and we're done */
	ret = pthread_mutex_unlock( &_Q(queue)->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
	}
	
	/* End of cleanup handler */
	return;
}

/* The internal function for meq_timedget and meq_get */
static int _meq_tget ( _meq_t * queue, msg_t ** msg, int istimed, const struct timespec *abstime)
{
	int ret = 0;
	int timedout = 0;
	_li_t * li;
	
	/* Check the parameters */
	if ( (! CHECK_QUEUE( queue )) || (msg == NULL) || ( istimed && (abstime == NULL)) ) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	/* Check the thread cancellation state */
	{
		int oldstate;
		ret = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_setcancelstate failed: %s", strerror(ret));
			return ret;
		}
		if (oldstate != PTHREAD_CANCEL_ENABLE) {
			(void) pthread_setcancelstate(oldstate, &ret);
			TRACE_DEBUG(INFO, "Invalid thread cancellation state to call this function");
			return EINVAL;
		}
	}
	/* Check the thread cancellation type */
	{
		int oldtype;
		ret = pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oldtype);
		if (ret != 0) {
			TRACE_DEBUG(INFO, "pthread_setcanceltype failed: %s", strerror(ret));
			return ret;
		}
		if (oldtype != PTHREAD_CANCEL_DEFERRED) {
			(void) pthread_setcanceltype(oldtype, &ret);
			TRACE_DEBUG(INFO, "Invalid thread cancellation type to call this function");
			return EINVAL;
		}
	}
	
	/* Initialize the msg value */
	*msg = NULL;
	
	/* lock the queue */
	ret = pthread_mutex_lock( &queue->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_lock failed: %s", strerror(ret));
		return ret;
	}
	
awaken:
	/* Check queue status */
	if (queue->count > 0) {
		/* There are messages in the queue, so pick the first one */
		LI_pop( &queue->list, li );
		queue->count--;
		*msg = li->msg;
		free(li);
	} else {
		/* We have to wait for a new message */
		queue->thrs++ ;
		pthread_cleanup_push(_meq_cleanup, queue);
		if (istimed) {
			ret = pthread_cond_timedwait( &queue->cond, &queue->mtx, abstime );
		} else {
			ret = pthread_cond_wait( &queue->cond, &queue->mtx );
		}
		pthread_cleanup_pop(0);
		queue->thrs-- ;
		if (ret == 0)
			goto awaken;  /* test for spurious wake-ups */
		
		if (istimed && (ret == ETIMEDOUT)) {
			timedout = 1;
		} else {
			/* Unexpected error condition */
			TRACE_DEBUG(INFO, "pthread_cond_wait or pthread_cond_timedwait failed: %s", strerror(ret));
			return ret;
		}
	}
	
	/* Unlock */
	ret = pthread_mutex_unlock( &queue->mtx );
	if (ret != 0) {
		TRACE_DEBUG(INFO, "pthread_mutex_unlock failed: %s", strerror(ret));
		return ret;
	}
	
	/* Done */
	return timedout ? ETIMEDOUT : 0;
}

	
/* Retrieve a message */
int meq_get ( meq_t * queue, msg_t ** msg )
{
	TRACE_ENTRY( "%p %p", queue, msg );
	return _meq_tget(_Q(queue), msg, 0, NULL);
}

/* Retrieve a message, if any. wait only a certain time */
int meq_timedget ( meq_t * queue, msg_t ** msg, const struct timespec *abstime )
{
	TRACE_ENTRY( "%p %p %p", queue, msg, abstime );
	return _meq_tget(_Q(queue), msg, 1, abstime);
}




